from django.conf import settings
from django.core.management import BaseCommand

from rq.cli import worker


class Command(BaseCommand):
    help = 'Starts an RQ worker.'

    def run_from_argv(self, argv):
        self.handle(
            f'--url={settings.REDIS_URL}',
            f'--job-class=rq_runner.job.AnsibleJob',
            *argv[2:]
        )

    def handle(self, *args, **options):
        worker(args)



import logging
from typing import Any, TYPE_CHECKING, Optional

from rq.job import Job
from rq.timeouts import BaseDeathPenalty
from rq.utils import import_attribute

if TYPE_CHECKING:
    from ansible_runner import Runner
    from redis.client import Pipeline
    from rq_result.backends import DatabaseBackend

from .plugins.callbacks.awx_display import MARK_EVENTS

logger = logging.getLogger(__name__)


class AnsibleJob(Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._backend = None
        self._backend_class = 'rq_result.backends.DatabaseBackend'
        self._runner: Optional['Runner'] = None

    @property
    def backend(self) -> 'DatabaseBackend':
        if self._backend is None:
            self._backend = import_attribute(self._backend_class)()
        return self._backend

    @property
    def runner(self) -> Optional['Runner']:
        return self._runner

    @runner.setter
    def runner(self, runner: Optional['Runner']):
        self._runner = runner

    def event_handler(self, event_data: dict[str, Any]) -> bool:
        try:
            if (
                    (ev := event_data.get('event')) in MARK_EVENTS and
                    (evd := event_data.get('event_data')) and
                    (custom_data := evd.get('custom_task_data'))
            ):
                self.meta['tasks'] = custom_data
                if ev == 'playbook_on_stats':
                    self.meta['stats'] = {'failures': evd.get('failures', {})}
                self.backend.store_result(self.id, meta=self.meta)
        except Exception as e:
            logger.error(f'Event handler error: {e}')

        return True

    def status_handler(self, status_data: dict[str, Any]):
        self.backend.store_result(self.id, **status_data)

    def _execute(self) -> Any:
        data = {
            'name': self.func_name,
            'meta': self.meta,
            'kwargs': self.kwargs,
            'description': self.description,
            'status': 'started',
            'queue': self.origin,
            'worker_name': self.worker_name or '',
            'group_id': self.group_id or '',
            'started_at': self.started_at,
            'enqueued_at': self.enqueued_at,
        }
        self.status_handler(data)
        return super()._execute()

    def _handle_success(self, result_ttl, pipeline: 'Pipeline', worker_name: str = ''):
        super()._handle_success(result_ttl, pipeline, worker_name)
        data = {
            'meta': self.meta,
            'status': 'finished',
            'ended_at': self.ended_at,
        }
        if not self._result is None:
            data['result'] = str(self._result)
        if not self._exc_info is None:
            data['exc_info'] = str(self._exc_info)
        if self.runner:
            data['rc'] = self.runner.rc
            data['status'] = self.runner.status
        self.status_handler(data)

    def _handle_failure(
            self,
            exc_string: str,
            pipeline: 'Pipeline',
            worker_name: str = ''
    ):
        super()._handle_failure(exc_string, pipeline, worker_name)
        data = {
            'meta': self.meta,
            'status': 'failed',
            'exc_info': exc_string,
            'ended_at': self.ended_at,
        }
        if not self._result is None:
            data['result'] = str(self._result)
        if self.runner:
            data['rc'] = self.runner.rc
        self.status_handler(data)

    def execute_stopped_callback(self, death_penalty_class: type[BaseDeathPenalty]):
        super().execute_stopped_callback(death_penalty_class)
        self.status_handler({
            'meta': self.meta,
            'status': 'stopped'
        })



from pathlib import Path


def get_settings_files(sub_dir='rq-runner', filename='settings.toml') -> list[Path]:
    return [
        Path.cwd() / filename,
        Path.home() / sub_dir / filename,
        Path(f'/etc/{sub_dir}/{filename}'),
    ]


def get_ansible_cfg() -> Path | None:
    for f in get_settings_files(filename='ansible.cfg'):
        if f.exists():
            return f
    return None



from __future__ import (absolute_import, division, print_function)

import contextlib
from collections import defaultdict
from datetime import datetime

from ansible.utils.color import colorize
from ansible_runner.display_callback.callback import awx_display

DOCUMENTATION = awx_display.DOCUMENTATION

MARK_EVENTS = [
    'playbook_on_task_start',
    'playbook_on_handler_task_start',
    'playbook_on_stats'
]


class CallbackModule(awx_display.CallbackModule):
    def __init__(self):
        super().__init__()
        self._host_total = 0
        self._prev_task = None
        self._task_data = defaultdict(dict)

    def _get_result_timing_data(self, result):
        try:
            self._task_data[result._task._uuid]['host_counter'] += 1
            result._result['progress'] = (
                f"{self._task_data[result._task._uuid]['host_counter']}/"
                f"{self._task_data[result._task._uuid]['host_total']}"
            )
        except Exception as e:
            self._display.error('Failed to set host progress: %s' % e)

        return super()._get_result_timing_data(result)

    def set_task(self, task, local=False):
        if not local:
            try:
                self._task_data[task._uuid]['name'] = task.get_name().strip()
                self._task_data[task._uuid]['host_counter'] = 0
                self._task_data[task._uuid]['host_total'] = self._host_total
                self._task_data[task._uuid]['started_at'] = datetime.now()
                self.set_task_duration()
                self._prev_task = task._uuid
            except Exception as e:
                self._display.error('Failed to set task: %s' % e)

        super().set_task(task, local)

    def set_task_duration(self):
        if not self._prev_task:
            return

        try:
            task = self._task_data[self._prev_task]
            task['ended_at'] = datetime.now()
            if not task['started_at']:
                return
            task['duration'] = (task['ended_at'] - task['started_at']).total_seconds()
        except Exception as e:
            self._display.error('Failed to set task duration: %s' % e)

    def set_options(self, task_keys=None, var_options=None, direct=None):
        setattr(self, '_load_name', awx_display.DefaultCallbackModule.CALLBACK_NAME)
        super().set_options(task_keys, var_options, direct)

    def v2_playbook_on_play_start(self, play):
        variable_manager = play.get_variable_manager()
        variable = variable_manager.get_vars(play=play)['vars']
        self._host_total = len(variable['ansible_play_hosts_all'])
        super().v2_playbook_on_play_start(play)

    def v2_playbook_on_stats(self, stats):
        self.set_task_duration()
        super().v2_playbook_on_stats(stats)

        if self._task_data:
            self._display.banner('TASK DATA')
            for v in self._task_data.values():
                if not v:
                    continue
                self._display.display(
                    u"%-46s : %-37s %s %s %s" % (
                        colorize(u'name', v['name'], awx_display.C.COLOR_OK),
                        colorize(
                            u'progress',
                            f"{v['host_counter']}/{v['host_total']}",
                            awx_display.C.COLOR_OK
                        ),
                        colorize(
                            u'started_at',
                            v.get('started_at'),
                            awx_display.C.COLOR_OK),
                        colorize(
                            u'ended_at',
                            v.get('ended_at', 'ended_at'),
                            awx_display.C.COLOR_OK
                        ),
                        colorize(
                            u'duration',
                            v.get('duration', 'NaN'),
                            awx_display.C.COLOR_OK
                        ),
                    ),
                    screen_only=True
                )

    @contextlib.contextmanager
    def capture_event_data(self, event, **event_data):
        if event in MARK_EVENTS:
            event_data['custom_task_data'] = [
                {k: str(v) for k, v in i.items()}
                for i in self._task_data.values()
            ]

        with super().capture_event_data(event, **event_data):
            yield



from pathlib import Path

from dynaconf import DjangoDynaconf

from .utils import get_settings_files

SECRET_KEY = 'django-insecure_rq-runner'
INSTALLED_APPS = ['rq_result', 'rq_runner']
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': Path.cwd() / 'db.sqlite3',
    }
}
TIME_ZONE = 'Asia/Shanghai'
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'
JOB_MODEL = 'rq_result.Job'
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(asctime)s | %(levelname)-8s | %(name)s - %(message)s',
        }
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
        },
        'file': {'class': 'logging.NullHandler', 'formatter': 'simple'},
    },
    'root': {
        'handlers': ['console', 'file'],
        'level': 'INFO',
        'propagate': False,
    },
}

settings = DjangoDynaconf(
    __name__,
    settings_file=get_settings_files(),
    environments=True,
    load_dotenv=False,
    merge_enabled=True
)

if settings.LOGFILE:
    LOGGING['handlers']['file'] = {
        'class': 'logging.handlers.TimedRotatingFileHandler',
        'filename': settings.LOGFILE,
        'level': 'INFO',
        'formatter': 'simple',
        'when': 'midnight',
        'backupCount': 14,
    }



import json
import logging
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Optional, TYPE_CHECKING

from ansible_runner import Runner, RunnerConfig
from ansible_runner.utils import dump_artifact, signal_handler
from rq import get_current_job

if TYPE_CHECKING:
    from rq.job import Job

from .utils import get_ansible_cfg

from .plugins import BASE_DIR

logger = logging.getLogger(__name__)


class AnsibleExecutor:
    def __init__(self):
        self._job = None
        self._event_handler = None
        self._status_handler = None
        self._artifacts_handler = None
        self._finished_callback = None
        self._env_vars = {
            'ANSIBLE_CONFIG': get_ansible_cfg(),
            'ANSIBLE_STDOUT_CALLBACK': 'default',
            'ANSIBLE_CALLBACK_FORMAT_PRETTY': True,
            'ANSIBLE_CALLBACK_RESULT_FORMAT': 'yaml',
            'AWX_LIB_DIRECTORY': str(BASE_DIR / 'callbacks'),
        }

    @property
    def job(self) -> Optional['Job']:
        if self._job is None:
            self._job = get_current_job()
        return self._job

    def event_handler(self, event_data: dict[str, Any]) -> bool:
        if self._job:
            self._job.event_handler(event_data)

        if self._event_handler:
            self._event_handler(event_data)

        return event_data.get('event') == 'playbook_on_stats'

    def status_handler(
            self,
            status_data: dict[str, Any],
            runner_config: 'RunnerConfig'
    ):
        if self._job and (status := status_data.get('status')) == 'running':
            self._job.status_handler({
                'status': status,
                'inventory': str(runner_config.inventory),
            })

        if self._status_handler is not None:
            self._status_handler(status_data, runner_config)

    def artifacts_handler(self, artifact_dir: str):
        if self._artifacts_handler is not None:
            self._artifacts_handler(artifact_dir)

    def finished_callback(self, runner: 'Runner'):
        try:
            if runner.stats and (failures := runner.stats.get('failures')):
                filename = Path(runner.config.artifact_dir) / 'retry_hosts'
                with open(filename, 'w') as f:
                    f.write('\n'.join(failures))
                logger.warning(f'To retry, use: --limit @{filename}\n')
        except Exception as e:
            logger.error('Save retry hosts failed: %s', e)

        if self._finished_callback is not None:
            self._finished_callback(runner)

    def __call__(self, *args, **kwargs):
        if self.job:
            kwargs['ident'] = self.job.id
        kwargs.setdefault('private_data_dir', None)
        kwargs.setdefault('suppress_env_files', True)
        kwargs['envvars'] = self._env_vars | kwargs.pop('envvars', {})
        for name in (
                'event_handler',
                'status_handler',
                'artifacts_handler',
                'finished_callback'
        ):
            setattr(self, f'_{name}', kwargs.pop(name, None))

        rc = RunnerConfig(**kwargs)
        if isinstance(rc.inventory, MutableMapping):
            rc.inventory = dump_artifact(
                obj=json.dumps(rc.inventory, indent=4),
                path=rc.artifact_dir,
                filename='hosts.json'
            )
        rc.prepare()
        r = Runner(
            rc,
            cancel_callback=signal_handler(),
            event_handler=self.event_handler,
            status_handler=self.status_handler,
            finished_callback=self.finished_callback
        )
        if self.job:
            setattr(self.job, 'runner', r)
            logger.info('Running job %s', str(self.job.id))

        return r.run()



from django.conf import settings
from django.core.management import BaseCommand
from dynaconf.utils.inspect import inspect_settings


class Command(BaseCommand):
    help = 'Debug settings.'

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        inspect_settings(settings, dumper='yaml', print_report=True)



from django.conf import settings
from django.test import TestCase
from redis import Redis
from rq import Queue

from .executor import AnsibleExecutor
from .job import AnsibleJob


class JobTestCase(TestCase):
    @staticmethod
    def adhoc():
        inventory = {
            'all': {
                'children': {
                    'test': {
                        'hosts': {
                            f'test_{i}': {
                                'plat': "test",
                                'sid': i,
                                'ansible_host': 'localhost',
                                'ansible_connection': 'local',
                            }
                            for i in range(3)
                        }
                    }
                },
                'vars': {}
            }
        }
        e = AnsibleExecutor()
        r = e(
            host_pattern='all',
            inventory=inventory,
            module='shell',
            module_args='ls -l /tmp',
            verbosity=2,
        )
        return r

    def test_job(self):
        conn = Redis.from_url(settings.REDIS_URL)
        q = Queue(
            'test',
            connection=conn,
            is_async=False,
            job_class=AnsibleJob,
        )
        job = q.enqueue(self.adhoc)
        print('job:', job)
        print('result:', job._result)
        print('exc_info:', job._exc_info)
